use super::*;

pub trait Inlining: Sized {
    fn iread_n<A, E>(
        self,
        n: usize,
        ok: impl FnOnce(&[u8]) -> A,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<(A, Self), E>;

    fn iread_all<A>(self, ok: impl FnOnce(&[u8]) -> A) -> A;

    fn itell(&self) -> usize;
}

pub trait InliningExt<E>: Inlining {
    fn iread_n_const<const N: usize>(
        self,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<([u8; N], Self), E> {
        self.iread_n(
            N,
            |slice| slice.try_into().expect("iread_n wrong length."),
            err,
        )
    }
}

impl<D: Inlining, E> InliningExt<E> for D {}

impl<D: ?Sized + Deserializer> Inlining for &mut D {
    fn iread_n<A, E>(
        self,
        n: usize,
        ok: impl FnOnce(&[u8]) -> A,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<(A, Self), E> {
        let slice = self.read_n(n);
        if slice.len() == n {
            Ok((ok(slice), self))
        } else {
            Err(err(slice))
        }
    }

    fn iread_all<A>(self, ok: impl FnOnce(&[u8]) -> A) -> A {
        ok(self.read_all())
    }

    fn itell(&self) -> usize {
        self.tell()
    }
}

pub trait InliningResultExt<A, E> {
    fn seal(self) -> Result<A, E>;
}

impl<A, E, I: Inlining> InliningResultExt<A, E> for Result<(A, I), E> {
    fn seal(self) -> Result<A, E> {
        self.map(|(a, _)| a)
    }
}
